#pragma once
#include <map>
#include <memory>
#include <optional>
#include <string>
#include <vector>
#include "instancetype.h"

#define INSTANCE_COPY_IMPL(CLASS_NAME) std::unique_ptr<Instance> CLASS_NAME::__copy() { return std::make_unique<CLASS_NAME>(CLASS_NAME(*this)); } 
#define INSTANCE_PRELUDE(CLASS_NAME) protected: virtual std::unique_ptr<Instance> __copy() override; virtual InstanceType& GetType() const override;
// #define INSTANCE_TYPE(CLASS_NAME, TYPE_DEFINITION) InstanceType& CLASS_NAME::GetType() const { \
//     static InstanceType type = TYPE_DEFINITION; \
//     return type; \
// };
#define INSTANCE_TYPE(CLASS_NAME, TYPE_CONSTANT) InstanceType& CLASS_NAME::GetType() const { return TYPE_CONSTANT; };

extern InstanceType TYPE_INSTANCE;

// The base class for all objects in the data model
class Instance : std::enable_shared_from_this<Instance> {
private:
    std::optional<std::weak_ptr<Instance>> Parent;
    std::vector<std::shared_ptr<Instance>> children;
protected:
    virtual std::unique_ptr<Instance> __copy();
public:
    std::string Name;
    std::optional<std::shared_ptr<Instance>> GetParent();
    void SetParent(std::optional<std::shared_ptr<Instance>> newParent);
    std::vector<std::shared_ptr<Instance>> GetChildren();

    virtual void Init();
    std::unique_ptr<Instance> CloneInternal();
    virtual InstanceType& GetType() const = 0;

    virtual Value GetProperty(std::string key);
    virtual void SetProperty(std::string key, Value newValue);
    virtual Value CallMethod(std::string key, std::vector<Value> arguments);
};